// This example is from the book "Java in a Nutshell, Second Edition".
// Written by David Flanagan.  Copyright (c) 1997 O'Reilly & Associates.
// You may distribute this source code for non-commercial purposes only.
// You may study, modify, and use this example for any purpose, as long as
// this notice is retained.  Note that this example is provided "as is",
// WITHOUT WARRANTY of any kind either expressed or implied.
package aima.agents;
import java.awt.*;               // ScrollPane, 
import java.awt.event.*;         // New event model.
import java.io.*;                // Object serialization streams.
import java.util.*;
import java.lang.Math;

/**
 * This class places a GridDisplay component in a ScrollPane container,
 * puts the ScrollPane in a window, and adds a simple pulldown menu system.
 * The menu uses menu shortcuts.  Events are handled with anonymous classes.
 */
public class GridFrame extends Frame {
  /** A very simple main() method for our program. */
  public static void main(String[] args) { new GridFrame(); }

  /** Remember # of open windows so we can quit when last one is closed */
  protected static int num_windows = 0;

  Label textLabel;

  /** Create a Frame, Menu, and ScrollPane for the gridDisplay component */
  public GridFrame() {
    super("GridFrame");                  // Create the window.
    num_windows++;                           // Count it.

    setLayout(new BorderLayout());
    Panel p = new Panel();
    p.add(new Button("Start"));
    p.add(new Button("Stop"));
    p.add(new Button("Restart"));
    p.add(textLabel = new Label("Here is some output info",java.awt.Label.LEFT));
    ScrollPane pane = new ScrollPane();      // Create a ScrollPane.
    pane.setSize(300, 300);                  // Specify its size.
    add(p,"North");
    add(pane,"Center");                // Add it to the frame.
    GridDisplay gridDisplay;
    gridDisplay = new GridDisplay(this, 500, 500); // Create a bigger gridDisplay area.
    pane.add(gridDisplay);                      // Add it to the ScrollPane.

    MenuBar menubar = new MenuBar();         // Create a menubar.
    this.setMenuBar(menubar);                // Add it to the frame.
    Menu file = new Menu("File");            // Create a File menu.
    menubar.add(file);                       // Add to menubar.

    // Create three menu items, with menu shortcuts, and add to the menu.
    MenuItem n, c, q;
    file.add(n = new MenuItem("New Window", new MenuShortcut(KeyEvent.VK_N)));
    file.add(c = new MenuItem("Close Window",new MenuShortcut(KeyEvent.VK_W)));
    file.addSeparator();                     // Put a separator in the menu
    file.add(q = new MenuItem("Quit", new MenuShortcut(KeyEvent.VK_Q)));

    // Create and register action listener objects for the three menu items.
    n.addActionListener(new ActionListener() {     // Open a new window
      public void actionPerformed(ActionEvent e) { new GridFrame(); }
    });
    c.addActionListener(new ActionListener() {     // Close this window.
      public void actionPerformed(ActionEvent e) { close(); }
    });
    q.addActionListener(new ActionListener() {     // Quit the program.
      public void actionPerformed(ActionEvent e) { System.exit(0); }
    });

    // Another event listener, this one to handle window close requests.
    this.addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) { close(); }
    });

    // Set the window size and pop it up.
    this.pack();
    this.show();


}

  /** Close a window.  If this is the last open window, just quit. */
  void close() {
    if (--num_windows == 0) System.exit(0);
    else this.dispose();
  }
}

/**
 * This class is a custom component that supports 2D-grids.  It also has
 * a popup menu that allows ?.
 * Note that it extends Component rather than Canvas, making it "lightweight."
 */
class GridDisplay extends Component implements ActionListener {
  protected short last_x, last_y;                // Coordinates of last click.
  protected Color current_color = Color.black;   // Current drawing color.
  protected int width, height;                   // The preferred size.
  protected PopupMenu popup;                     // The popup menu.
  protected GridFrame frame;                         // The frame we are within.

  /** This constructor requires a Frame and a desired size */
  public GridDisplay(GridFrame frame, int width, int height) {
    this.frame = frame;
    this.width = width;
    this.height = height;

    // We handle scribbling with low-level events, so we must specify
    // which events we are interested in.
    this.enableEvents(AWTEvent.MOUSE_EVENT_MASK);
    this.enableEvents(AWTEvent.MOUSE_MOTION_EVENT_MASK);

    // Create the popup menu using a loop.  Note the separation of menu
    // "action command" string from menu label.  Good for internationalization.
    String[] labels   = new String[] {"Clear" };
    String[] commands = new String[] {"clear"};
    popup = new PopupMenu();                   // Create the menu
    for(int i = 0; i < labels.length; i++) {
      MenuItem mi = new MenuItem(labels[i]);   // Create a menu item.
      mi.setActionCommand(commands[i]);        // Set its action command.
      mi.addActionListener(this);              // And its action listener.
      popup.add(mi);                           // Add item to the popup menu.
    }
    Menu colors = new Menu("Color");           // Create a submenu.
    popup.add(colors);                         // And add it to the popup.
    String[] colornames = new String[] { "Black", "Red", "Green", "Blue"};
    for(int i = 0; i < colornames.length; i++) {
      MenuItem mi = new MenuItem(colornames[i]);  // Create the submenu items
      mi.setActionCommand(colornames[i]);         // in the same way.
      mi.addActionListener(this);
      colors.add(mi);
    }
    // Finally, register the popup menu with the component it appears over
    this.add(popup);
  }

  /** Specifies how big the component would like to be.  It always returns the
   *  preferred size passed to the GridDisplay() constructor */
  public Dimension getPreferredSize() { return new Dimension(width, height); }

  /** This is the ActionListener method invoked by the popup menu items */
  public void actionPerformed(ActionEvent event) {
    // Get the "action command" of the event, and dispatch based on that.
    // This method calls a lot of the interesting methods in this class.
    String command = event.getActionCommand();
    if (command.equals("clear")) clear();
    else if (command.equals("Black")) current_color = Color.black;
    else if (command.equals("Red")) current_color = Color.red;
    else if (command.equals("Green")) current_color = Color.green;
    else if (command.equals("Blue")) current_color = Color.blue;
  }

  int xSize = 10, ySize = 10, boxSize = 30, minBoxSize = 20;
  /** Draw all the saved lines of the gridDisplay, in the appropriate colors */
  public void paint(Graphics g) {
    // Draw the grid lines
    int xWin = this.getSize().width, yWin = this.getSize().height;
    boxSize = java.lang.Math.max(minBoxSize, java.lang.Math.min(xWin/xSize, yWin/ySize));
    for(int i = 0; i <= xSize; i++) {
      g.drawLine(i*boxSize, 0, i*boxSize, ySize*boxSize);
      g.drawLine(0, i*boxSize,  xSize*boxSize, i*boxSize);
    }
  }

  /**
   * This is the low-level event-handling method called on mouse events
   * that do not involve mouse motion.  Note the use of isPopupTrigger()
   * to check for the platform-dependent popup menu posting event, and of
   * the show() method to make the popup visible.  If the menu is not posted,
   * then this method saves the coordinates of a mouse click or invokes
   * the superclass method.
   */
  public void processMouseEvent(MouseEvent e) {
    if (e.isPopupTrigger())                               // If popup trigger,
      popup.show(this, e.getX(), e.getY());               // pop up the menu.
    else if (e.getID() == MouseEvent.MOUSE_PRESSED) {
      last_x = (short)e.getX(); last_y = (short)e.getY(); // Save position.
    }
    else super.processMouseEvent(e);  // Pass other event types on.
  }

  /**
   * This method is called for mouse motion events.  It adds a line to the
   * gridDisplay, on screen, and in the saved representation
   */
  public void processMouseMotionEvent(MouseEvent e) {

      frame.textLabel.setText(e.getX()/boxSize + "," + e.getY()/boxSize);
   super.processMouseMotionEvent(e);  // Important!
  }

  /** Clear the gridDisplay.  Invoked by popup menu */
  void clear() {
    repaint();                   // and redraw everything.
  }
}
