package aima.cs.learning;

import java.util.*;
import java.io.*;
import java.awt.*;
import aima.math.Function;

/**
  * A basic Q-learning agent.
  *
  * Both states and actions are represented and communicated using ints.
  *
  * @author Daishi Harada (daishi@cs.berkeley.edu)
  */
public class QLearning_Agent
  extends Canvas
  implements Serializable, StateReward_PerceptListener
{
  public QLearning_Agent() {
    randomSource = new Random();
  }

  public QLearning_Agent(int n, int na) {
    this();
    numStates = n;
    numActions = na;
    allocParams();
  }

  /**
    * Sets the current state of the agent.
    * The assumption is that this state is detached from the
    * previous state history.
    */
  public void setState(int s) {
    state = s;
    prevState = -1;
  }

  public int getState() {
    return (state);
  }

  public void setNumStates(int n) {
    numStates = n;
    allocParams();
  }

  public int getNumStates() {
    return (numStates);
  }

  public void setNumActions(int na) {
    numActions = na;
    allocParams();
  }

  public int getNumActions() {
    return (numActions);
  }
  
  protected void allocParams() {
    if (numStates != -1 && numActions != -1) {
      Q = new double[numStates][numActions];
      V = new double[numStates];
      policy = new int[numStates];
    }
  }

  public void randomlyInitParams(double max) {
    int i,j;
    for (i=0;i<numStates;i++) {
      for (j=0;j<numActions;j++) {
	Q[i][j] = max*randomSource.nextDouble();
      }
    }
    computeValues();
    computePolicy();
  }

  protected void computeValues() {
    int i;
    for (i=0;i<numStates;i++) {
      computeValueOf(i);
    }
  }

  protected void computeValueOf(int i) {
    V[i] = Function.max(Q[i]);
  }

  protected void computePolicy() {
    int i;
    for (i=0;i<numStates;i++) {
      computePolicyOf(i);
    }
  }

  protected void computePolicyOf(int i) {
    policy[i] = Function.argmax(Q[i]);
  }

  public void setLearningRate(double r) {
    learningRate = r;
  }

  public double getLearningRate() {
    return (learningRate);
  }

  public void setExplorationRate(double r) {
    explorationRate = r;
  }

  public double getExplorationRate() {
    return (explorationRate);
  }

  public void setDiscountFactor(double d) {
    discountFactor = d;
  }

  public double getDiscountFactor() {
    return (discountFactor);
  }

  /**
    * Take a experience/learning step.
    * Input is the new state and reward. Returns an action.
    * Assume that input state of -1 indicates reset
    * (that the previous state was terminal.)
    */
  public int step(int s, double r) {
    if (verbosity > 2) {
      logStream.println("Q-learnstep: new state "+s+" prevsta "+state+ " rew "+r);
      logStream.println(this.toString());
    }
    if (s == -1) {
      V[state] = 0.0;
      prevAction = -1;
      state = -1;
    } else {      
      prevState = state;
      state = s;
      if (prevState != -1) {
	Q[prevState][prevAction] +=
	  learningRate * (discountFactor*(r+V[state])-Q[prevState][prevAction]);
	computeValueOf(prevState);
	computePolicyOf(prevState);
      }
      prevAction = getAction();
    }
    if (verbosity > 3) {
      logStream.println(this.toString());
    }
    return (prevAction);
 } 

  /**
    * Returns an action.
    * This method will need to be modified when more sophisticated
    * exploration policies are desired.
    */
  public int getAction() {
    int a;
    if (randomSource.nextDouble() < explorationRate) {
      a = (int)(numActions*randomSource.nextDouble());
    } else {
      a = policy[state];
    }
    return (a);
  }

  public String toString() {
    int i,j;
    StringBuffer b = new StringBuffer();
    b.append(this.getClass().getName());
    b.append("\n");
    for (i=0;i<numStates;i++) {
      for(j=0;j<numActions;j++) {
	b.append(Q[i][j]+" ");
      }
      b.append("\n");
    }
    return (b.toString());
  }

  private double learningRate;
  private double explorationRate;
  private double discountFactor;
  private int state = -1;
  private int prevState = -1;
  private int prevAction = -1;
  private int numStates = -1;
  private int numActions = -1;
  private double[][] Q;
  private double[] V;
  private int[] policy;
  private Random randomSource;

  public void setVerbosity(int n) {
    verbosity = n;
  }

  public int getVerbosity() {
    return (verbosity);
  }

  private int verbosity = 0;
  private PrintStream logStream = System.err;

  //-----------------------------------------------------------------
  // Bean code follows:
  // (most of this is taken verbatim from the BDK example bean OurButton.)
  //-----------------------------------------------------------------

  public void perceiveStateReward(StateReward_PerceptEvent e) {
    int action = step(e.getState(), e.getReward());
    if (action >= 0)
      {
	Vector targets;
	synchronized (this) {
	  targets = (Vector) pushListeners.clone();
	}
	intActionEvent actionEvt = new intActionEvent(this, action);
	for (int i = 0; i < targets.size(); i++) {
	  intActionListener target = (intActionListener)targets.elementAt(i);
	  target.intActionPerformed(actionEvt);
	}
      }
  }
  
  /**
    * The specified ActionListeners actionPerformed method will 
    * be called each time the button is clicked.  The ActionListener
    * object is added to a list of ActionListeners managed by 
    * this button, it can be removed with removeActionListener.
    * Note: the JavaBeans specification does not require ActionListeners
    * to run in any particular order.
    * 
    * @see #removeActionListener
    * @param l the ActionListener
    */      
  public synchronized void addIntActionListener(intActionListener l) {
    pushListeners.addElement(l);
  }
  
  /** 
    * Remove this ActionListener from the buttons internal list.  If the
    * ActionListener isn't on the list, silently do nothing.
    * 
    * @see #addActionListener
    * @param l the ActionListener
    */      
  public synchronized void removeIntActionListener(intActionListener l) {
    pushListeners.removeElement(l);
  }

  public synchronized void paint(Graphics g) {
    int height = 20;
    int width = 50;
    g.setColor(Color.lightGray);
    g.fill3DRect(0, 0, width - 1, height - 1, true);
    g.setColor(Color.black);
    g.setFont(new Font("Dialog", Font.PLAIN, 12));
    g.drawRect(2, 2, width - 4, height - 4);
    FontMetrics fm = g.getFontMetrics();
    g.drawString(label, (width - fm.stringWidth(label)) / 2, 
		 (height + fm.getMaxAscent() - fm.getMaxDescent()) / 2);
  }

  private String label = "Q-agent";

  public Dimension getMinimumSize() {
    FontMetrics fm = getFontMetrics(getFont());
    return new Dimension(fm.stringWidth(label) + TEXT_XPAD, 
			 fm.getMaxAscent() + fm.getMaxDescent() + TEXT_YPAD);
  }

  private Vector pushListeners = new Vector();
  static final int TEXT_XPAD = 12;
  static final int TEXT_YPAD = 8;
}

