package aima.stat.rv;

import java.io.*;
import java.util.*;

/**
  * An abstract utility class to be subclassed by finite RV's.
  *
  * This class provides the basic utility functions needed by
  * simple implementations of finite RV's where sample values
  * are indexed by {0,1,...,n-1}.
  *
  * @author Daishi Harada (daishi@cs.berkeley.edu)
  */
public abstract class FiniteSampler
{
  /**
    * Constructor initializing only the random source.
    */
  public FiniteSampler() {
    randomSource = new Random();
  }

  /**
    * Constructor initializing the random source and the number of sample values.
    *
    * @param n The number of sample values.
    */
  public FiniteSampler(int n) {
    this();
    this.n = n;
    prob = new double[n];
  }

  /**
    * Constructor initializing the random source, the number of sample values, and the distribution.
    *
    * @param n The number of sample values.
    * @param p[] The distribution over the sample values.
    * p[i] = probability that <code>indexSample()</code> returns i.
    * @see FiniteSampler#indexSample
    */
  public FiniteSampler(int n, double[] p) {
    this(n);
    int i;
    for (i=0;i<n;i++) {
      prob[i] = p[i];
    }
  }

  /**
    * Sets the number of elements.
    */
  public void setN(int n) {
    this.n = n;
  }

  /**
    * Gets the number of elements.
    */
  public int getN() {
    return (n);
  }

  /**
    * Sets the distribution. 
    */
  public void setProb(double[] p) {
    int i;
    for (i=0;i<n;i++) {
      prob[i] = p[i];
    }
  }

  /**
    * Gets the distribution. 
    */
  public double[] getProb() {
    return (prob);
  }

  /**
    * Sets the random source.
    */
  public void setRandomSource(Random randomSource) {
    this.randomSource = randomSource;
  }

  /**
    * Gets the random source.
    */
  public Random getRandomSource() {
    return (randomSource);
  }

  /**
    * Normalizes the distribution.
    * Does not do any error checking, possible to cause divide-by-zero.
    */
  public void normalize() {
    int i;
    double tmp = 0.0;
    for (i=0;i<n;i++) {
      tmp += prob[i];
    }
    for (i=0;i<n;i++) {
      prob[i] /= tmp;
    }
  }

  /**
    * Returns one of {0,1,...,n-1} following the distribution.
    * Simply walks through the distribution, and hence running time is
    * linear in n. In theory, this could be dropped to log(n).
    *
    * @return i with probability prob[i] (0 <= i < n.)
    */
  protected int indexSample() {
    int i = 0;
    double cum = prob[0];
    double tmp = randomSource.nextDouble();
    if (verbosity > 2) {
      log("tmp="+tmp+", cum="+cum);
    }
    while (tmp > cum) {
      cum += prob[++i];
      if (verbosity > 2) {
	log("tmp="+tmp+", cum="+cum);
      }
    }
    if (verbosity > 2) {
      log("return="+i);
    }
    return (i);
  }
  
  private void log(String message) {
    if (verbosity > 0) {
      logStream.println("["+this.getClass().getName()+"] "+message);
    }
  }

  private int n;
  private double[] prob;
  private Random randomSource;

  private int verbosity = 0;
  private PrintStream logStream = System.err;
}
