package aima.util;

import java.io.*;
import java.util.*;
import aima.*;

/**
  * The Base Environment.
  *
  * All AIMA Environments will subclass from this class
  */
public abstract class BaseEnvironment
  implements Environment
{
  public BaseEnvironment() {
  }

  public BaseEnvironment(Agent[] agents) {
    this.agents = new VectorSet();
    for (int i=0;i<agents.length;i++) {
      addAgent(agents[i]);
    }
  }

  public void addAgent(Agent agent) {
    agents.add(agent);
  }

  public void removeAgent(Agent agent) {
    agents.remove(agent);
  }
	
  public void setNumStepsToRun(int numStepsToRun) {
    this.numStepsToRun = numStepsToRun;
  }

  public int getNumStepsToRun() {
    return (numStepsToRun);
  }

  /**
    * Run the environment for numSteps time steps.
    */
  public BaseEnvironment run(int numStepsToRun) {
    setNumStepsToRun(numStepsToRun);
    return (run());
  }

  /**
    * The main run() method for Environments.
    */
  public BaseEnvironment run() {
    int i;
    Enumeration agentEnum;
    Agent agent;
    Percept percept;
    Action action;
    for (i=0;i<numStepsToRun && !termination(); i++, numSteps++) {
      log("*** Step: " + numSteps + "; "+agents.size()+" agents");
      // Give Agents their percepts.
      agentEnum = agents.iterator();
      while (agentEnum.hasNext()) {
	agent = (Agent)agentEnum.next();
	percept = getAgentPercept(agent);
	log("Agent "+agent+" saw "+percept);
	agentManager.setAgentPercept(agent, percept);
      }
      // Get Actions from agents, execute their actions.
      // Will need modification if we want agent actions
      // to occur synchronously. (alternatively, the actions
      // themselves could be coded to ensure synchronization).
      agentEnum = agents.elements();
      while (agentEnum.hasNext()) {
	agent = (Agent)agentEnum.next();
	action = agentManager.getAgentAction(agent);
	log("Agent "+agent+" did "+action);
	setAgentAction(agent, action);
      }
      if (synchronous) {
	executeSynchronously();
      }
      // Compute agent's performance
      // Not included yet. This needs to be thought about
      // so that it interfaces well with both printing, dumping
      // to a separate DataSet datastructure for further manipulation, etc.
    }
    return this;
  }

  /**
    * Gets the current percept for agent.
    */
  protected abstract Percept getAgentPercept(Agent agent);

  /**
    * Sets the action returned by the agent.
    * Unless the environment is synchronous, this should also
    * execute the action.
    */
  protected abstract void setAgentAction(Agent agent, Action action);

  /**
    * Execute all agent's actions synchronously.
    * Synchronous environments will have to have stored the agent's actions
    * using setAgentAction.
    */
  protected void executeSynchronously() {
  }

  /**
    * 
    */
  protected Set getActions() {
    return (new Set());
  }

  protected Set getPercepts() {
    return (new Set());
  }

  /**
    * @returns true if the simulation should end now.
    */
  protected boolean termination() {
    return (false);
  }

  public String toString () {
    return "[" + name + "]";
  }

  protected static String name = "BaseEnvironment";

  /**
    * Set holding the agents in this Environment.
    */
  protected Set agents;

  /**
    * Set of the Environment's actions.
    */
  protected Set actions;

  /**
    * Set of the Environment's percepts.
    */
  protected Set percepts;

  /**
    * The agent manager
    */
  protected AgentManager agentManager;

  /**
    * Number of steps run so far.
    */
  protected int numSteps = 0;

  /**
    * Number of steps to run per call to run()
    */
  protected int numStepsToRun = 10;

  /**
    * Flag saying whether the environment needs to
    * be updated synchronously.
    */
  protected boolean synchronous = false;

  /**
    * Log function
    */
  protected void log(String message) {
    if (verbosity > 0) {
      logStream.println("["+this+"]"+message);
    }
  }

  /**
    * The verbosity of this Environment
    */
  protected int verbosity = 0;

  /**
    * PrintStream to which to write log messages.
    */
  protected PrintStream logStream = System.err;
}




