package aima.util;

import aima.*;

public class GridEnvironment
  extends BaseEnvironment
{
  public GridEnvironment(Agent[] agents, int xSize, int ySize) {
    super(agents);
    actions = getActions();
    percepts = getPercepts();
    this.xSize = xSize; this.ySize = ySize;
    grid = new Vector[xSize][ySize];
    for (int x = 0; x<xSize; x++)
      for (int y = 0; y<ySize; y++)
        grid[x][y] = new Vector();
    for (int i=0; i<agents.length; i++)
      placeObject(agents[i], agents[i].location);
  }

  protected Set getActions() {
    actions = super.getActions();
    actions.add(new TurnAction());
    actions.add(new ForwardAction());
    actions.add(new GrabAction());
    actions.add(new ReleaseAction());
    actions.add(new SpeakAction());
    return (actions);
  }

  protected Set getPercepts() {
    percepts = super.getPercepts();
    percepts.add(new BumpPercept());
    return (percepts);
  }

  /**
    * Since GridEnvironment provide Agent.Bodies, we need
    * to give each agent a body when they're added.
    */
  public void addAgent(Agent agent) {
    super.addAgent(agent);
    agent.setBody(new GridAgentBody());
  }

  public void setAgentAction(Agent agent, Action action) {
    if (actions.contains(action)) {
      action.execute();
    } else {
      // signal error
    }
  }

  /** Put walls all around the perimeter of the environment. */
  public GridEnvironment addWalls() {
    for (int x = 0; x < xSize; x++) {
      addWall(x, 0); addWall(x, ySize-2);
    }
    for (int y =  1; y < ySize-1; y++) {
      addWall(0, y); addWall(xSize - 1, y);
    }
    return this;
  }
	
  private void addWall(int x, int y) {
    placeObject(new Wall(), new Point(x,y));
  }
  
  public void addWithProbability(Class objectClass, float prob) {
    for (int x = 1; x<xSize-1; x++)
      for (int y = 1; y<ySize-1; y++)
        if (Math.random() < prob)
          try {
            placeObject((Thing)objectClass.newInstance(), new Point(x,y));
          } catch (Exception e) {
            System.exit(1);
          }
  }
	
  /** Find a point somewhere in the environment that satisfies test.
   * After TRIES, give up and return null. */
  private Point randomLoc(UnaryPredicate test, int tries) {
    for (int i = 0; i < tries; i++) {
      Point loc = new Point(randomInt(xSize), randomInt(ySize));
      if (test.execute(loc)) {return loc;}
    }
    return null;
  }
	
  /* return a random integer from 0..n-1, uniformly distributed */
  int randomInt(int n) {
    return (int)java.lang.Math.round(java.lang.Math.random() * n) / n;
  }
  private boolean isFreeLoc (Point loc) {
    return !loc.equals(start) && findObjectOfType(Obstacle.class, loc) == null;
  }

	public Vector gridContents(int x, int y) {
	  return grid[x][y];
	}
	
  public Vector gridContents(Point loc) { // ??? contained?
    return grid[loc.x][loc.y];
  }
	
  //// Dealing with Objects
	
  /** Move a thing to a location, and return that location.
   * If there's an obstacle there, return nil and give object a bump. */
  public Point moveObjectTo(Thing object, Point loc) {
    if (findObjectOfType(Obstacle.class, loc) != null) {
      object.isBump = true;
      return null;
    } else {
      removeObject(object);
      placeObject(object, loc);
      return loc;
    }
  }

  /** Place an object at this location. */
  public void placeObject(Thing object, Point loc) {
    object.location = loc;
    gridContents(loc).addElement(object);
    p("Placed " + object + " at " + loc);
    if (step == 0) { objects.addElement(object); }
  }

  public void placeInContainer(Thing object, Thing container) {
    removeObject(object);
    container.contents.addElement(object);
    object.location = container.location;
  }
  /* Remove this object from this location. */
  public void removeObject(Thing object) {
    gridContents(object.location).removeElement(object);
  }

  /** Is there an object of this type here? */
  protected Thing findObjectOfType(Class type, Point loc) {
    Vector objs = gridContents(loc);
    for(int i=0; i<objs.size(); i++) {
      if (type.isInstance(objs.elementAt(i))) return (Thing)objs.elementAt(i);
    }
    return null;
  }
  
  private int xSize;
  private int ySize;
  private Vector[][] grid;
  public Point start = new Point(1, 1);
  private Vector objects = new Vector();

  public class Thing
    implements Environment.Thing
  {
    public String  name = "?";
    public boolean isAlive = false;
    public Point   location;
    public boolean isBump = false;
    public float   size = 0.5f;
    public Color   color = Color.black;
    public String  shape = "rectangle";
    public String  sound = null;
    public Vector  contents = new Vector();
    public Thing   holding = null;
    public Thing   containedIn = null;
    public Point   heading = new Point(0,0);
  }

  public interface Obstacle
  {
  }

  public class Wall
    extends Thing implements Obstacle
  {
  }

  public class AgentBody
    extends Thing implements Agent.Body
  {
  }

  public class TurnAction
    extends BaseAction
  {
    public TurnAction() {
      super("turn");
    }
    public void execute() {
    }
  }
}
