package aima.util;

import aima.*;
import aima.cs.util.*;

/**
  * This is the default AgentManager for Environments.
  * It creates a new thread for each agent.
  * Threads are {\em not} assigned to particular agents.
  * Rather, threads are woken up as percepts come in, first come
  * first serve. However, since there are as many threads as there are
  * agents, there should in principle be no contention (assuming the
  * environment does not obliviously dump percepts at the agents.)
  */
public class ThreadedAgentManager
  implements Environment.AgentManager, Runnable
{
  /**
    * Adding a new agent creates a new thread for the agent.
    */
  public void addAgent(Agent agent) {
    (new Thread(this)).start();
  }
  
  /**
    * Should GC threads. noop for now.
    */
  public void removeAgent(Agent agent) {
  }
  
  /**
    * Sends percept to agent.
    * Uses the queue as the lock.
    * Notifies a waiting thread of new agent/percept.
    */
  public void setAgentPercept(Agent agent, Percept percept) {
    synchronized(perceptQueue) {
      perceptQueue.push(new AgentPercept(agent,percept));
      synchronized(actionStaleSet) {
	actionStaleSet.add(agent);
      }
      perceptQueue.notify();
    }
  }
  
  /**
    * Gets the current action taken by the agent.
    * Note that this does not check whether this action is in response
    * to the last percept sent to the agent.
    */
  public Action getAgentAction(Agent agent) {
    return ((Action)actionTable.get(agent));
  }
  
  /**
    * Gets the 
    */
  public Action getFreshAgentAction(Agent agent)
    throws InterruptedException {
    synchronized(actionStaleSet) {
      while (actionStaleSet.contains(agent)) {
	actionStaleSet.wait();
      }
      return (getAgentAction(agent));
    }
  }
  
  /**
    */
  public void receiveAction(ActionEvent actionEvent) {
    setAgentAction((Agent)actionEvent.getSource(), (Action)actionEvent.getAction());
  }
  
  /**
    * Main thread body.
    * All of the wait()ing happens in getAgentPercept()
    */
  public void run() {
    try {
      while (true) {
	AgentPercept tmp = getAgentPercept();
	Agent agent = tmp.agent;
	PerceptEvent perceptEvent = new PerceptEvent(this, tmp.percept);
	agent.execute(perceptEvent);
      }
    } catch (InterruptedException e) {
    }
  }
  
  /**
    * Gets the next AgentPercept.
    * Uses the queue as lock.
    */
  protected AgentPercept getAgentPercept()
    throws InterruptedException {
    synchronized(perceptQueue) {
      while (perceptQueue.isEmpty()) {
	perceptQueue.wait();
      }
    }
    return ((AgentPercept)perceptQueue.pop());
  }

  /**
    * Updates agent's action.
    */
  protected void setAgentAction(Agent agent, Action action) {
    synchronized(actionStaleSet) {
      synchronized(actionTable) {
	actionTable.put(agent, action);
      }
      actionStaleSet.remove(agent);
      actionStaleSet.notify();
    }
  }

  protected Queue perceptQueue;
  protected Table actionTable;
  protected Set actionStaleSet;

  /**
    * Utility class for ThreadedAgentManager.
    * Packages agents and percepts together.
    */
  private class AgentPercept
  {
    public AgentPercept(Agent agent, Percept percept) {
      this.agent = agent;
      this.percept = percept;
    }
    public Agent agent;
    public Percept percept;
  }
}
